#include <gtk/gtk.h>
#include <cls.h>
#include <filtermenu.h>
#include <heap_dump.h>
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <jmp-config.h>

static gint olist_selected_row = -1;

void olist_row_selected (GtkWidget *widget,
			 gint row,
			 gint column,
			 GdkEventButton *event,
			 gpointer data) {
    olist_selected_row = row;
}

void olist_row_unselected (GtkWidget *widget,
			   gint row,
			   gint column,
			   GdkEventButton *event,
			   gpointer data) {
    olist_selected_row = -1;
}

/** request to show all live objects by a certain class. */
static void omenu_show_alloced (GtkObject *object_list) {
    cls* c = gtk_clist_get_row_data (GTK_CLIST (object_list), olist_selected_row);
    run_show_objects_alloced_by_class (c);
}

/** request to show all owners of a given class type. */
static void omenu_find_owners (GtkObject *object_list) {
    cls* c = gtk_clist_get_row_data (GTK_CLIST (object_list), olist_selected_row);
    run_find_owners (c);
}


/** Build a menu to show the options avialable on a class.
 */
static GtkWidget *build_omenu (GtkWidget* object_list) {
    cls* c;
    GtkWidget *menuitem;
    GtkWidget *filters;
    GtkWidget *omenu = gtk_menu_new ();
    menuitem = gtk_menu_item_new_with_label (_("show alloc'ed instances"));
    gtk_menu_append (GTK_MENU (omenu), menuitem);
    gtk_widget_show (menuitem);
    gtk_signal_connect_object (GTK_OBJECT (menuitem), "activate",
			       GTK_SIGNAL_FUNC (omenu_show_alloced),
			       GTK_OBJECT (object_list));
    

    /** todo only grey this menu entry out.. */
    menuitem = gtk_menu_item_new_with_label (_("Show object owners"));
    if (!is_get_owners_possible ()) 
	gtk_widget_set_sensitive (menuitem, FALSE);

    gtk_menu_append (GTK_MENU (omenu), menuitem);
    gtk_widget_show (menuitem);
    gtk_signal_connect_object (GTK_OBJECT (menuitem), "activate",
			       GTK_SIGNAL_FUNC (omenu_find_owners),
			       GTK_OBJECT (object_list));
    
    /* add filters... */
    c = gtk_clist_get_row_data (GTK_CLIST (object_list), olist_selected_row);
    filters = build_filter_menu (c);
    menuitem = gtk_menu_item_new_with_label (_("Filter"));
    gtk_menu_append (GTK_MENU (omenu), menuitem);
    gtk_widget_show (menuitem);
    gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), filters);
    return omenu;
}

gint olist_button_handler (GtkWidget *widget,
			   GdkEventButton *event,
			   gpointer  callback_data) {
    if (event->button == 3 && olist_selected_row != -1) {
	GtkWidget* omenu = build_omenu (widget);
	gtk_menu_popup (GTK_MENU (omenu), NULL, NULL, NULL, NULL,
                        event->button, event->time);	
	return TRUE;
    }
    return FALSE;
}
