#ifndef HEAP_DUMP_H__
#define HEAP_DUMP_H__

#include <jni.h>
#include <jvmpi.h>

/** Handle the heap dumps. */

/** or these values together to get the four (4) combinations possible.. */
#define STATIC_VARIABLE 0
#define VARIABLE 1
#define ARRAY 2

typedef struct object_link object_link;

struct object_link {
    jobjectID  obj;    /** The object in question. */
    jobjectID  parent; /** The parent/owner of obj. */
    jobjectID  clz;    /** The class of the parent. 
			   Needed to find right variable. */
    object_link* next; /** The next owner of of this obj. */ 
    int        type;   /** Type of link. Or value of VARIABLE and ARRAY */
    
    jint       variable; /** index in field array for class. */
    jint       index;    /** array index. */
};


/** Handle the heap dump level 0. */
void heap_dump_0 (int dump_level,
		  char *begin,
		  char *end,
		  jint num_traces,
		  JVMPI_CallTrace *traces);

/** Handle the heap dump level 1. */
void heap_dump_1 (int dump_level,
		  char *begin,
		  char *end,
		  jint num_traces,
		  JVMPI_CallTrace *traces);

/** Handle the heap dump level 2. */
void heap_dump_2 (int dump_level,
		  char *begin,
		  char *end,
		  jint num_traces,
		  JVMPI_CallTrace *traces);


void object_dump (jint data_len, char *data); 

/** Get the owners of the given object at the last heap dump.
 * @param obj the object to get the owners for.
 * @return an linked list of object_link containing the owners of obj.
 */
object_link* get_owners (jobjectID obj);


/** Check if it is possible to get object owners...
 */
int is_get_owners_possible ();

#endif /* HEAP_DUMP_H__ */
