#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <cls.h>
#include <obj.h>
#include <ui.h>
#include <heap_dump.h>
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <jmp-config.h>

static char buf[1024];

typedef struct tc {
    GtkWidget* tree;
    cls* clz;
} tc;

static int not_in_stack (jobjectID obj, jobjectID* stack, size_t level) {
    size_t i;
    for (i = 0; i < level; i++) 
	if (obj == stack[i])
	    return 0;
    return 1;
}

static void list_owners (jobjectID obj_id, size_t level, 
			 size_t maxlevel, jobjectID* stack, 
			 GtkWidget* tree) {
    GtkWidget *item;
    GtkWidget *subtree;
    object_link* owners;
    if (level >= maxlevel)
	return;
    stack[level] = obj_id;
    owners = get_owners (obj_id);
    while (owners != NULL) {
	obj* parent;
	cls* pc;
	cls* vc;
	char* var = NULL;
	parent = get_object (owners->parent);
	pc = get_class (obj_get_class_id (parent));
	vc = get_class (owners->clz);
	switch (owners->type) {	    
	case STATIC_VARIABLE:
	    var = vc->statics[owners->variable].field_name;
	    snprintf (buf, 1024, _("%p is a static variable (%s) in class %s"),
		      owners->obj, var, vc->name);
	    break;
	case VARIABLE:
	    var = vc->instances[owners->variable].field_name;
	    snprintf (buf, 1024, _("%p is a variable (%s) in object %p of class %s"),
		      owners->obj, var, owners->parent, pc->name);
	    break;
	case ARRAY:
	    var = "[]";
	    snprintf (buf, 1024, _("%p is located in an array %p of %s at index %d"),
		      owners->obj, owners->parent, vc->name, owners->index);
	    break;
	}

	item = gtk_tree_item_new_with_label (buf);
	gtk_tree_append (GTK_TREE(tree), item);
	gtk_widget_show (item);

	if (not_in_stack (owners->parent, stack, level) && 
	    owners->parent != owners->obj &&
	    !owners->type == STATIC_VARIABLE) {
	    subtree = gtk_tree_new();
	    gtk_tree_set_view_mode (GTK_TREE(subtree), GTK_TREE_VIEW_ITEM);
	    gtk_tree_item_set_subtree (GTK_TREE_ITEM(item), subtree);
	    list_owners (owners->parent, level + 1, maxlevel, stack, subtree);
	}
	owners = owners->next;
    }
}

static void find_instance_owners (void* object, void* tz) {
    obj* o = (obj*)object;
    tc* t = (tc*)tz;
    cls* c = t->clz;
    size_t maxlevel = 7;
    jobjectID* save = malloc (sizeof (jobjectID) * maxlevel);
    if (obj_get_class_id (o) == c->class_id) {
	list_owners (o->obj_id, 0, maxlevel, save, t->tree);
    }
}

void show_instance_owners (hashtab* objects, cls* c) {
    GtkWidget *win;
    GtkWidget *scrolledwindow1;
    GtkWidget *tree;    
    GtkWidget *subtree;
    GtkWidget *item;
    tc tz;

    win = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_object_set_data (GTK_OBJECT (win), "win", win);
    gtk_window_set_title (GTK_WINDOW (win), _("Instance owners"));
    scrolledwindow1 = gtk_scrolled_window_new (NULL, NULL);
    gtk_widget_ref (scrolledwindow1);
    gtk_object_set_data_full (GTK_OBJECT (win), "scrolledwindow1", scrolledwindow1,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (scrolledwindow1);
    gtk_container_add (GTK_CONTAINER (win), scrolledwindow1);
    tree = gtk_tree_new();
    gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW(scrolledwindow1),
					   tree);
    gtk_tree_set_selection_mode (GTK_TREE(tree),
				 GTK_SELECTION_SINGLE);
    gtk_widget_show (tree);
    
    snprintf (buf, 1024, "%s", cls_get_name (c));
    item = gtk_tree_item_new_with_label (buf);
    gtk_tree_append (GTK_TREE(tree), item);
    gtk_widget_show (item);
    subtree = gtk_tree_new();
    gtk_tree_set_view_mode (GTK_TREE(subtree), GTK_TREE_VIEW_ITEM);
    gtk_tree_item_set_subtree (GTK_TREE_ITEM(item), subtree);

    tz.clz = c;
    tz.tree = subtree;
    hash_for_each_with_arg (find_instance_owners, objects, &tz);
    
    gtk_widget_set_usize (win, 500, 200);
    gtk_widget_show (win);
}
