#ifndef JMP_H__
#define JMP_H__

#include <jni.h>
#include <jvmpi.h>
#include <hash.h>

/** These are standard type of objects, they are defined in there
 * respective header files.
 */
typedef struct cls cls;
typedef struct obj obj;
typedef struct method method;
typedef struct methodtime methodtime;
typedef struct jthread jthread;
typedef struct arena arena;
typedef struct timerstack timerstack;
typedef struct mvector mvector;

/** Get the real object from its id. */
cls*        get_class (const jobjectID class_id);
hashtab*    get_classes ();
obj*        get_object (const jobjectID obj_id);
method*     get_method (const jmethodID method_id);
hashtab*    get_methods ();
jthread*    get_jthread (const JNIEnv* env_id);

/** Get the timerstack for a thread. */
timerstack* get_timerstack (JNIEnv* env_id);

/** Get all the threads */
hashtab*    get_threads ();

/** Get the lock on the object monitor. */
void lock_objects ();

/**  Unlock the object monitor. */
void unlock_objects ();

/** Get the lock on the thread monitor. */
void lock_threads ();

/** Unlock the thread monitor. */
void unlock_threads ();

/** returns true if we are currently tracing objects. */
int tracing_objects ();

/** returns true if we are currently tracing methods. */
int tracing_methods ();

/** Test if jmp is using its GUI */
int usingUI ();

/** Reset all current counters... */
void reset_counters ();

/** Restore all counters... */
void restore_counters ();

/** Run the java garbage collector. */
void run_GC ();

/** Run the heap dump. */
void run_heap_dump ();

/** request the object info... */
void get_object_alloc (jobjectID obj_id);

/** request the object info... */
void get_class_load (jobjectID class_id);

/** request the class. */
void get_class_load (jobjectID class_id);

/** request that jmp dumps it data. 
 * @return the id of the dump file.
 */
int run_data_dump ();

/** request that jmp dumps objects alloced in the given method. 
 * @param m the method that has alloced objects.
 */
void run_show_objects_alloced_by_method (method* m);

/** request that jmp dumps objects by a given class
 * @param c the class of the object to dump.
 */
void run_show_objects_alloced_by_class (cls* c);

/** request that jmp show object owners for all instances of a given class.
 * @param c the class of the instaces jmp should find owners for.
 */
void run_find_owners (cls* c);

/** Get the super class of the specified class. 
 *  This will also fill in the cls->super pointer..
 * @param c the cls to get the super class of.
 * @return the super class of c 
 */
cls* get_super_class (cls* c);


void check_objects (size_t cnt);

#endif /* JMP_H__ */
