#include <string.h>
#include <gtk/gtk.h>
#include <filter.h>
#include <cls.h>
#include <method.h>
#include <mvector.h>
#include <method_info.h>
#include <filtermenu.h>
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <jmp-config.h>

static gint mlist_selected_row = -1;
static char buf[1024];

void mlist_row_selected (GtkWidget *widget,
			 gint row,
			 gint column,
			 GdkEventButton *event,
			 gpointer data) {
    mlist_selected_row = row;
}

void mlist_row_unselected (GtkWidget *widget,
			   gint row,
			   gint column,
			   GdkEventButton *event,
			   gpointer data) {
    mlist_selected_row = -1;
}

/** request to show all live objects allocated by a method. */
static void mmenu_show_alloced (GtkObject *method_list) {
    method* m = gtk_clist_get_row_data (GTK_CLIST (method_list), mlist_selected_row);
    run_show_objects_alloced_by_method (m);
}

/** Add a subtree to the call graph. 
 * @param m the method we are adding to.
 * @param tree the tree widget.
 * @param depth a counter so we dont recurse infinatly 
 *  (a call graph is not a tree and may have circular refs).
 */
static void add_method_tree (method* m, GtkWidget *tree, size_t depth) {
    size_t i;
    GtkWidget *item;
    GtkWidget *subtree;
    mvector *mv = m->called_methods;

    if (depth >= 10)
	return;
    for (i = 0; i < mvector_load (mv); i++) {
	method* mm = mvector_get (mv, i);
	snprintf (buf, 1024, "%s: %s", mm->owner->name, mm->jmpname);
	item = gtk_tree_item_new_with_label (buf);
	gtk_tree_append (GTK_TREE(tree), item);
	gtk_widget_show (item);	

	if (mvector_load (mm->called_methods)) {
	    subtree = gtk_tree_new();
	    gtk_tree_set_view_mode (GTK_TREE(subtree), GTK_TREE_VIEW_ITEM);
	    gtk_tree_item_set_subtree (GTK_TREE_ITEM(item), subtree);
	    add_method_tree (mm, subtree, depth + 1);
	}
    }
}

/** Create a window with a call graph for the currently selected method. */
static void mmenu_show_called (GtkObject *method_list) {
    GtkWidget *win;
    GtkWidget *scrolledwindow1;
    GtkWidget *tree;    
    GtkWidget *subtree;
    GtkWidget *item;
    method* m = gtk_clist_get_row_data (GTK_CLIST (method_list), mlist_selected_row);
    win = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_object_set_data (GTK_OBJECT (win), "win", win);
    gtk_window_set_title (GTK_WINDOW (win), _("Call graph"));
    scrolledwindow1 = gtk_scrolled_window_new (NULL, NULL);
    gtk_widget_ref (scrolledwindow1);
    gtk_object_set_data_full (GTK_OBJECT (win), "scrolledwindow1", scrolledwindow1,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (scrolledwindow1);
    gtk_container_add (GTK_CONTAINER (win), scrolledwindow1);
    tree = gtk_tree_new();
    gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW(scrolledwindow1),
					   tree);
    gtk_tree_set_selection_mode (GTK_TREE(tree),
				 GTK_SELECTION_SINGLE);
    gtk_widget_show (tree);

    snprintf (buf, 1024, "%s: %s", m->owner->name, m->jmpname);
    item = gtk_tree_item_new_with_label (buf);
    gtk_tree_append (GTK_TREE(tree), item);
    gtk_widget_show (item);
    subtree = gtk_tree_new();
    gtk_tree_set_view_mode (GTK_TREE(subtree), GTK_TREE_VIEW_ITEM);
    gtk_tree_item_set_subtree (GTK_TREE_ITEM(item), subtree);

    add_method_tree (m, subtree, 0);
    
    gtk_widget_set_usize (win, 500, 200);
    gtk_widget_show (win);
}

static void mmenu_show_method_info (GtkObject *method_list) {
    method* m = gtk_clist_get_row_data (GTK_CLIST (method_list), mlist_selected_row);
    show_method_info (m);
}


static GtkWidget* build_mmenu (GtkWidget* method_list) {
    GtkWidget* menuitem;
    GtkWidget* mmenu;
    GtkWidget* filters;
    method* m;
    
    mmenu = gtk_menu_new ();
    menuitem = gtk_menu_item_new_with_label (_("show alloc'ed instances"));
    gtk_menu_append (GTK_MENU (mmenu), menuitem);	
    gtk_widget_show (menuitem);
    gtk_signal_connect_object (GTK_OBJECT (menuitem), "activate",
			       GTK_SIGNAL_FUNC (mmenu_show_alloced),
			       GTK_OBJECT (method_list));
    
    menuitem = gtk_menu_item_new_with_label (_("show called methods"));
    gtk_menu_append (GTK_MENU (mmenu), menuitem);	
    gtk_widget_show (menuitem);
    gtk_signal_connect_object (GTK_OBJECT (menuitem), "activate",
			       GTK_SIGNAL_FUNC (mmenu_show_called),
			       GTK_OBJECT (method_list));
    
    menuitem = gtk_menu_item_new_with_label (_("show method info"));
    gtk_menu_append (GTK_MENU (mmenu), menuitem);
    gtk_widget_show (menuitem);
    gtk_signal_connect_object (GTK_OBJECT (menuitem), "activate",
			       GTK_SIGNAL_FUNC (mmenu_show_method_info),
			       GTK_OBJECT (method_list));

    /* add filters... */
    m = gtk_clist_get_row_data (GTK_CLIST (method_list), mlist_selected_row);
    filters = build_filter_menu (m->owner);
    menuitem = gtk_menu_item_new_with_label (_("Filter"));
    gtk_menu_append (GTK_MENU (mmenu), menuitem);
    gtk_widget_show (menuitem);
    gtk_menu_item_set_submenu (GTK_MENU_ITEM (menuitem), filters);

    return mmenu;
}

gint mlist_button_handler (GtkWidget *widget,
			   GdkEventButton  *event,
			   gpointer   callback_data) {
    if (event->button == 3 && mlist_selected_row != -1) {
	GtkWidget* mmenu = build_mmenu (widget);
	gtk_menu_popup (GTK_MENU (mmenu), NULL, NULL, NULL, NULL,
                        event->button, event->time);	
	return TRUE;
    }
    return FALSE;
}

