#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <cls.h>
#include <hash.h>
#include <jmp.h>
#include <ui.h>
#include <method.h>
#include <cls.h>
#include <obj.h>
#include <jthread.h>
#include <mvector.h>
#include <filter.h>
#include <methodtime.h>
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <jmp-config.h>

static char* Q = "?";
static char* SYS_INIT = "[system init]";
static char sarena[64];
static char is_array[64];
static char size[64];

typedef struct {
    GtkWidget *window;
    GtkWidget *clist;
    gint       row;
    void      *data;
} wret;

static void append_object_dump (obj* o, wret* w) {
    gchar *text[6];
    cls* c = NULL;
    cls* cp = get_class (obj_get_class_id (o));

    /* titles = { "Class name", "arena", "is array", "size", "allocating class", "allocating method"};*/    
    method* m = o->method;
    if (m != NULL)
	c = m->owner;
    text[0] = (cp ? cp->name : Q);
    snprintf (sarena, 64, "%d", o->arena_id);
    text[1] = sarena;
    snprintf (is_array, 64, "%d", (o->is_array ? 1 : 0));
    text[2] = is_array;
    snprintf (size, 64, "%d", o->size);
    text[3] = size;
    text[4] = (c ? c->name : Q);
    text[5] = (m ? m->jmpname : SYS_INIT);
    gtk_clist_append (GTK_CLIST (w->clist), text);
    gtk_clist_set_row_data (GTK_CLIST (w->clist), w->row++, o);
}

static void print_if_method (obj* o, wret* w) {
    if (o->method == (method*)w->data)
	append_object_dump (o, w);
}

static void print_if_class (obj* o, wret* w) {
    if (obj_get_class_id (o) == ((cls*)w->data)->class_id)
	append_object_dump (o, w);
}

static gint sort_by_class_name (GtkCList     *clist,
				gconstpointer ptr1,
				gconstpointer ptr2) {
    obj* opp1 = (obj*)(GTK_CLIST_ROW ((GtkCListRow*)ptr1));
    obj* opp2 = (obj*)(GTK_CLIST_ROW ((GtkCListRow*)ptr2));
    cls* cp1 = get_class (obj_get_class_id (opp1));
    cls* cp2 = get_class (obj_get_class_id (opp2));
    char* cn1 = cp1 ? cp1->name : Q;
    char* cn2 = cp2 ? cp2->name : Q;
    return (strcmp (cn1, cn2));    
}

static gint sort_by_arena (GtkCList     *clist,
			   gconstpointer ptr1,
			   gconstpointer ptr2) {
    obj* o1 = (obj*)(GTK_CLIST_ROW ((GtkCListRow*)ptr1));
    obj* o2 = (obj*)(GTK_CLIST_ROW ((GtkCListRow*)ptr2));
    return o1->arena_id - o2->arena_id;
}

static gint sort_by_is_array (GtkCList     *clist,
			      gconstpointer ptr1,
			      gconstpointer ptr2) {
    obj* o1 = (obj*)(GTK_CLIST_ROW ((GtkCListRow*)ptr1));
    obj* o2 = (obj*)(GTK_CLIST_ROW ((GtkCListRow*)ptr2));
    return o1->is_array - o2->is_array;
}

static gint sort_by_size (GtkCList     *clist,
			  gconstpointer ptr1,
			  gconstpointer ptr2) {
    obj* o1 = (obj*)(GTK_CLIST_ROW ((GtkCListRow*)ptr1));
    obj* o2 = (obj*)(GTK_CLIST_ROW ((GtkCListRow*)ptr2));
    return o1->size - o2->size;
}

static gint sort_by_alloc_class_and_method (GtkCList     *clist,
					    gconstpointer ptr1,
					    gconstpointer ptr2) {
    cls* c1 = NULL;
    cls* c2 = NULL;
    char* cn1;
    char* cn2;
    int i;
    obj* o1 = (obj*)(GTK_CLIST_ROW ((GtkCListRow*)ptr1));
    obj* o2 = (obj*)(GTK_CLIST_ROW ((GtkCListRow*)ptr2));
    method *m1 = o1->method;
    method *m2 = o2->method;
    if (m1)
	c1 = m1->owner;
    if (m2)
	c2 = m2->owner;    
    cn1 = c1 ? c1->name : Q;
    cn2 = c2 ? c2->name : Q;
    i = strcmp (cn1, cn2);
    if (i)
	return i;
    cn1 = m1 ? m1->jmpname : SYS_INIT;
    cn2 = m2 ? m2->jmpname : SYS_INIT;
    i = strcmp (cn1, cn2);
    return i;
}

static int ((*obj_comprs[])(GtkCList *clist, 
			    gconstpointer ptr1,
			    gconstpointer ptr2)) = { sort_by_class_name,
						     sort_by_arena, 
						     sort_by_is_array,          /* is_array */
						     sort_by_size,              /* size */
						     sort_by_alloc_class_and_method,  /* alloc_class */
						     sort_by_alloc_class_and_method   /* alloc_method */
			    };


static void obj_column_clicked (GtkWidget *widget, gint column) {
    if (obj_comprs[column]) {
	gtk_clist_set_compare_func (GTK_CLIST (widget),
				    (GtkCListCompareFunc)obj_comprs[column]);
	gtk_clist_sort (GTK_CLIST (widget));
    } else {
	fprintf (stdout, _("Sort order not yet implemented.\n"));
    }
}

wret create_object_dump_window () {
    GtkWidget *obj_dump;
    GtkWidget *scrolledwindow1;
    GtkWidget *clist1;
    wret ret;
    gchar *titles[] = { _("Class name"), _("Arena"), _("Is array"), _("Size"), 
			_("Allocating class"), _("Allocating method")};
    
    obj_dump = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_object_set_data (GTK_OBJECT (obj_dump), "obj_dump", obj_dump);
    gtk_window_set_title (GTK_WINDOW (obj_dump), _("Object dump"));
    scrolledwindow1 = gtk_scrolled_window_new (NULL, NULL);
    gtk_widget_ref (scrolledwindow1);
    gtk_object_set_data_full (GTK_OBJECT (obj_dump), "scrolledwindow1", scrolledwindow1,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (scrolledwindow1);
    gtk_container_add (GTK_CONTAINER (obj_dump), scrolledwindow1);
    
    clist1 = gtk_clist_new_with_titles (6, titles);
    gtk_widget_ref (clist1);
    gtk_object_set_data_full (GTK_OBJECT (obj_dump), "clist1", clist1,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (clist1);
    gtk_clist_column_titles_show (GTK_CLIST (clist1));
    gtk_container_add (GTK_CONTAINER (scrolledwindow1), clist1);
    
    gtk_clist_set_column_width (GTK_CLIST (clist1), 0, 200);
    gtk_clist_set_column_width (GTK_CLIST (clist1), 4, 200);
    gtk_clist_set_column_width (GTK_CLIST (clist1), 5, 200);

    gtk_signal_connect(GTK_OBJECT(clist1), "click_column", 
		       GTK_SIGNAL_FUNC(obj_column_clicked),
		       NULL);    
    
    
    gtk_widget_set_usize (obj_dump, 800, 200);
    ret.window = obj_dump;
    ret.clist = clist1;
    return ret;
}

void show_objects_alloced_by_method (hashtab* objects, method* m) {
    wret w = create_object_dump_window ();
    w.data = m;
    w.row = 0;
    hash_for_each_with_arg ((hash_iter_fa)print_if_method, objects, &w);
    gtk_widget_show (w.window);
}

void show_objects_alloced_by_class (hashtab* objects, cls* c) {
    wret w = create_object_dump_window ();
    w.data = c;
    w.row = 0;
    hash_for_each_with_arg ((hash_iter_fa)print_if_class, objects, &w);
    gtk_widget_show (w.window);
}
