#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <gtk/gtk.h>
#include <jni.h>
#include <jmp-config.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <hash.h>
#include <method.h>
#include <jthread.h>
#include <methodtime.h>
#include <threads_window.h>
#include <ui.h>

static gint threads_selected_row = -1;
static size_t current_row = 0;
static size_t num_threads = 0;
static jthread** threadlist = NULL;

/** compare threads based on thread name */
static int jthread_compr_name (const void* v1, const void* v2) {
    jthread** t1 = (jthread**)v1;
    jthread** t2 = (jthread**)v2;
    return (strcmp (jthread_get_thread_name (*t1), jthread_get_thread_name (*t2)));
}

/** compare threads based on thread group */
static int jthread_compr_group (const void* v1, const void* v2) {
    jthread** t1 = (jthread**)v1;
    jthread** t2 = (jthread**)v2;
    return (strcmp (jthread_get_group_name (*t1), jthread_get_group_name (*t2)));
}

/** compare threads based on thread parent */
static int jthread_compr_parent (const void* v1, const void* v2) {
    jthread** t1 = (jthread**)v1;
    jthread** t2 = (jthread**)v2;
    return (strcmp (jthread_get_parent_name (*t1), jthread_get_parent_name (*t2)));
}

/** Compare methods based on used time in the method. */
static int jthread_compr_contenation (const void* v1, const void* v2) {
    long l;
    jthread** t1 = (jthread**)v1;
    jthread** t2 = (jthread**)v2;
    if ((l = ((*t1)->contended.tv_sec - (*t2)->contended.tv_sec)) < 0) {
	return 1;
    } else if (l > 0) {
	return 0;
    } else {
	return (*t2)->contended.tv_usec - (*t1)->contended.tv_usec;
    }
}


/** array of the thread compare functions... */
static int ((*jthread_comprs[])(const void* v1, const void* v2)) = { jthread_compr_name, 
								     jthread_compr_group, 
								     jthread_compr_parent,
								     jthread_compr_contenation };
								     
/** our current comparator */
static int (*jthread_compr) (const void* v1, const void* v2) = jthread_compr_name;

static void threads_column_clicked (GtkWidget *widget, gint column) {
    if (jthread_comprs[column] != NULL) {
	if (jthread_compr != jthread_comprs[column]) {
	    jthread_compr = jthread_comprs[column];
	}
    } else {
	fprintf (stdout, "Sort order not yet implemented.\n");
    }
}

static void add_thread (void* thread, void* clist, int row_num) {
    char* row[4];
    char buf[32];
    jthread* jt = (jthread*)thread;
    
    row[0] = jthread_get_thread_name (jt);
    row[1] = jthread_get_group_name (jt);
    row[2] = jthread_get_parent_name (jt);
    snprintf (buf, 32, "%ld.%06ld", jt->contended.tv_sec, jt->contended.tv_usec);
    row[3] = buf;
    if (row_num < num_threads) {
	int j;
	for (j = 0; j < 3; j++) 
	    gtk_clist_set_text (GTK_CLIST (clist), row_num, j, row[j]);
    } else {
	gtk_clist_append (GTK_CLIST (clist), row);
    }
    gtk_clist_set_row_data (GTK_CLIST (clist), current_row++, jt);
}

void threads_row_selected (GtkWidget *widget,
			   gint row,
			   gint column,
			   GdkEventButton *event,
			   gpointer data) {
    GtkWidget* stacklist = (GtkWidget*)data;
    char* crow[2];
    int i;
    jthread* t;
    threads_selected_row = row;
    
    gtk_clist_freeze (GTK_CLIST (stacklist));
    t = gtk_clist_get_row_data (GTK_CLIST (widget), threads_selected_row);
    if (t != NULL) {
	timerstack* ts = get_timerstack (jthread_get_env_id (t)); 
	if (ts != NULL) {
	    for (i = ts->top - 1; i >= 0; i--) {
		method* m;
		methodtime* mt = ts->times + i;
		m = get_method (mt->method_id);
		crow[0] = (m ? cls_get_name (method_get_owner (m)) : "???");
		crow[1] = (m ? method_get_method_jmpname (m) : "???");
		gtk_clist_append (GTK_CLIST (stacklist), crow);	
	    }
	}
    }
    gtk_clist_thaw (GTK_CLIST (stacklist));
}

void threads_row_unselected (GtkWidget *widget,
			     gint row,
			     gint column,
			     GdkEventButton *event,
			     gpointer data) {
    GtkWidget* stacklist = (GtkWidget*)data;
    threads_selected_row = -1;
    gtk_clist_freeze (GTK_CLIST (stacklist));
    gtk_clist_clear (GTK_CLIST (stacklist));
    gtk_clist_thaw (GTK_CLIST (stacklist));
}

static gint delete_event (GtkWidget *widget, GdkEvent  *event, gpointer   data) {
    gtk_widget_hide (widget);
    return (TRUE);
}

static void count_threads (void* thread, void* counter) {
    int* c = (int*)counter;
    (*c)++;
}

static void add_thread1 (void* thread, void* tl) {
    jthread** threadlist = (jthread**)tl;
    threadlist[current_row++] = (jthread*)thread;
}

static void update_threads_table (GtkWidget* threads_table, hashtab* threads) {
    int counter = 0;
    int i;   
    gtk_clist_freeze (GTK_CLIST (threads_table));
    
    hash_for_each_with_arg ((hash_iter_fa)count_threads, threads, &counter);
    if (counter != num_threads)
	threadlist = realloc (threadlist, sizeof (jthread*) * counter);
    current_row = 0;
    hash_for_each_with_arg ((hash_iter_fa)add_thread1, threads, threadlist);
    qsort (threadlist, counter, sizeof (jthread*), jthread_compr);
    current_row = 0;
    for (i = 0; i < counter; i++) 
	add_thread (threadlist[i], threads_table, i);
    if (counter < num_threads) {
	for (i = num_threads - 1; i >= counter; i--)
	    gtk_clist_remove (GTK_CLIST (threads_table), i);
    }
    num_threads = counter;
    gtk_clist_thaw (GTK_CLIST (threads_table));
}

threads_window* create_threads_window (hashtab* threads) {
    GtkWidget *window2;
    GtkWidget *vbox2;
    GtkWidget *vbox3;
    GtkWidget *scrolledwindow1;
    GtkWidget *scrolledwindow2;
    GtkWidget *threadslabel;
    GtkWidget *clist1;
    GtkWidget *stacklabel;
    GtkWidget *tstack;
    GtkWidget *hpane;
    gchar *titles[] = { _("Name"), _("Group"), _("Parent"), _("Contenation")};
    gchar *stacktitles[] = { _("Class"), _("Method")};
    threads_window *tw;

    window2 = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_signal_connect (GTK_OBJECT (window2), "delete_event",
			GTK_SIGNAL_FUNC (delete_event), NULL);
    gtk_object_set_data (GTK_OBJECT (window2), "window2", window2);
    gtk_widget_set_usize (window2, 900, 200);
    gtk_window_set_title (GTK_WINDOW (window2), _("Current threads"));

    scrolledwindow1 = gtk_scrolled_window_new (NULL, NULL);
    gtk_widget_ref (scrolledwindow1);
    gtk_object_set_data_full (GTK_OBJECT (window2), "scrolledwindow1", scrolledwindow1,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (scrolledwindow1);

    threadslabel = gtk_label_new (_("Threads"));
    gtk_widget_ref (threadslabel);
    gtk_object_set_data_full (GTK_OBJECT (window2), "threadslabel", threadslabel,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (threadslabel);

    clist1 = gtk_clist_new_with_titles (4, titles);
    gtk_widget_ref (clist1);
    gtk_object_set_data_full (GTK_OBJECT (window2), "clist1", clist1,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (clist1);

    vbox2 = gtk_vbox_new (FALSE, 0);
    gtk_widget_ref (vbox2);
    gtk_object_set_data_full (GTK_OBJECT (window2), "vbox2", vbox2,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (vbox2);
    gtk_container_add (GTK_CONTAINER (scrolledwindow1), clist1);
    gtk_clist_set_column_width (GTK_CLIST (clist1), 0, 175);
    gtk_clist_set_column_width (GTK_CLIST (clist1), 1, 80);
    gtk_clist_set_column_width (GTK_CLIST (clist1), 2, 80);
    gtk_clist_set_column_width (GTK_CLIST (clist1), 3, 50);
    gtk_clist_column_titles_show (GTK_CLIST (clist1));
    
    gtk_box_pack_start (GTK_BOX (vbox2), threadslabel, FALSE, FALSE, 0);
    gtk_box_pack_start (GTK_BOX (vbox2), scrolledwindow1, TRUE, TRUE, 0);

    
    stacklabel = gtk_label_new (_("Stack"));
    gtk_widget_ref (stacklabel);
    gtk_object_set_data_full (GTK_OBJECT (window2), "stacklabel", stacklabel,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (stacklabel);

    tstack = gtk_clist_new_with_titles (2, stacktitles);
    gtk_widget_ref (tstack);
    gtk_object_set_data_full (GTK_OBJECT (window2), "tstack", tstack,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (tstack);
    gtk_clist_set_column_width (GTK_CLIST (tstack), 0, 180);
    gtk_clist_set_column_width (GTK_CLIST (tstack), 1, 180);

    scrolledwindow2 = gtk_scrolled_window_new (NULL, NULL);
    gtk_widget_ref (scrolledwindow2);
    gtk_object_set_data_full (GTK_OBJECT (window2), "scrolledwindow2", scrolledwindow2,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (scrolledwindow2);
    gtk_container_add (GTK_CONTAINER (scrolledwindow2), tstack);
    
    vbox3 = gtk_vbox_new (FALSE, 0);
    gtk_widget_ref (vbox3);
    gtk_object_set_data_full (GTK_OBJECT (window2), "vbox3", vbox3,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (vbox3);

    gtk_box_pack_start (GTK_BOX (vbox3), stacklabel, FALSE, FALSE, 0);
    gtk_box_pack_start (GTK_BOX (vbox3), scrolledwindow2, TRUE, TRUE, 0);

    hpane = gtk_hpaned_new ();
    gtk_widget_ref (hpane);
    gtk_object_set_data_full (GTK_OBJECT (window2), "hpane", hpane,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (hpane);
    gtk_container_add (GTK_CONTAINER (window2), hpane);

    gtk_widget_show (hpane);
    gtk_paned_add1 (GTK_PANED (hpane), vbox2);
    gtk_paned_add2 (GTK_PANED (hpane), vbox3);
    gtk_paned_set_position (GTK_PANED (hpane), 450);
    
    gtk_signal_connect (GTK_OBJECT (clist1), "click_column", 
			GTK_SIGNAL_FUNC (threads_column_clicked),
			NULL);    
    gtk_signal_connect (GTK_OBJECT (clist1),
			"select_row",
			GTK_SIGNAL_FUNC (threads_row_selected),
			tstack);
    gtk_signal_connect (GTK_OBJECT (clist1),
			"unselect_row",
			GTK_SIGNAL_FUNC (threads_row_unselected),
			tstack);
    
    update_threads_table (clist1, threads);

    tw = malloc (sizeof (threads_window));
    tw->window = window2;
    tw->threads_table = clist1;
    return tw;
}

void update_threads_window (threads_window* tw, hashtab* threads) {
    update_threads_table (tw->threads_table, threads);
}
