#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <cls.h>
#include <hash.h>
#include <jmp.h>
#include <ui.h>
#include <method.h>
#include <cls.h>
#include <obj.h>
#include <jthread.h>
#include <methodtime.h>
#include <filter.h>
#include <classlist_menu.h>
#include <methodlist_menu.h>
#include <threads_window.h>
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <jmp-config.h>

/** The widgets for the object window. */
static GtkWidget *jmpwin;
static GtkWidget *scroller;
static GtkWidget *object_list;
static int object_list_size = 0;

/** The widgets for the method window. */
static GtkWidget *mwin;
static GtkWidget *mscroller;
static GtkWidget *mlist;
static int mlist_size = 0;

/** The main (button) window  */
static GtkWidget *JMPmain;
static GtkWidget *statusbar;
static GtkWidget *about;

static GtkTooltips *tooltips;

/** The window of the current threads.. */
static threads_window *threadswin = NULL;

/** state tokens. */
static int        init_done = 0;
static int        quit = 0;
static int        is_updating = 1;

/** arrays that hold the methods/objects we show. */
static cls** classlist = NULL;
static jthread** threadslist = NULL;
static method** methodlist = NULL;
static int classes_count = 0;
static int threads_count = 0;
static int methods_count = 0;

/** id of last dump file... */
static int dump_id = -1;

/** Are we dumping on a timer basis? */
static long  dumptimer = 0;   /* The countdown value (10)... */
static long  dumptime = 0;    /* Current value (1, 2, 3, 4,....) */

/** A class to show total numbers in class info window. */
static cls* total;

/** compare classes based on class name */
static int cls_compr_name (const void* v1, const void* v2) {
    cls** c1;
    cls** c2;
    c1 = (cls**)v1;
    c2 = (cls**)v2;
    return (strcmp (cls_get_name (*c1), cls_get_name (*c2)));
}

/** compare classes based on instance count */
static int cls_compr_instance (const void* v1, const void* v2) {
    cls** c1;
    cls** c2;
    c1 = (cls**)v1;
    c2 = (cls**)v2;
    return (cls_get_instances (*c2) - cls_get_instances (*c1));
}

/** compare classes based on maximum instance count */
static int cls_compr_max_instance (const void* v1, const void* v2) {
    cls** c1;
    cls** c2;
    c1 = (cls**)v1;
    c2 = (cls**)v2;
    return (cls_get_max_instances (*c2) - cls_get_max_instances (*c1));
}

/** compare classes based on current size in bytes */
static int cls_compr_size (const void* v1, const void* v2) {
    cls** c1;
    cls** c2;
    c1 = (cls**)v1;
    c2 = (cls**)v2;
    return (cls_get_size (*c2) - cls_get_size (*c1));
}

/** array of the class compare functions... */
static int ((*cls_comprs[])(const void* v1, const void* v2)) = { cls_compr_name, 
								 cls_compr_instance, 
								 cls_compr_max_instance, 
								 cls_compr_size };

/** our current comparator */
static int (*cls_compr) (const void* v1, const void* v2) = cls_compr_size;


/** method comparators */


/** Compare methods based on signature. */
static int method_compr_signature (const void* v1, const void* v2) {
    method** m1;
    method** m2;
    m1 = (method**)v1;
    m2 = (method**)v2;
    return (strcmp (method_get_method_signature (*m1), method_get_method_signature (*m2)));
}

/** Compare methods based on method name (and signature if needed) */
static int method_compr_name (const void* v1, const void* v2) {
    method** m1;
    method** m2;
    int c = 0; 
    m1 = (method**)v1;
    m2 = (method**)v2;
    c = (strcmp (method_get_method_name (*m1), method_get_method_name (*m2)));
    if (c) 
	return c;
    return method_compr_signature (v1, v2);
}

/** Compare method based on class name, method name and signature (in that order) */
static int method_compr_class (const void* v1, const void* v2) {
    method** m1;
    method** m2;
    cls* c1;
    cls* c2;
    int c;
    m1 = (method**)v1;
    m2 = (method**)v2;
    c1 = method_get_owner (*m1);
    c2 = method_get_owner (*m2);
    c = (strcmp (cls_get_class_name (c1), cls_get_class_name (c2)));    
    if (c)
	return c;
    return method_compr_name (v1, v2);
}

/** Compare methods based on used time in the method. */
static int method_compr_time (const void* v1, const void* v2) {
    method** m1;
    method** m2;
    long l;
    m1 = (method**)v1;
    m2 = (method**)v2;
    if ((l = ((*m1)->time_used.tv_sec - (*m2)->time_used.tv_sec)) < 0) {
	return 1;
    } else if (l > 0) {
	return 0;
    } else {
	return (*m2)->time_used.tv_usec - (*m1)->time_used.tv_usec;
    }
}

/** Compare methods based on number of calls. */
static int method_compr_calls (const void* v1, const void* v2) {
    method** m1;
    method** m2;
    m1 = (method**)v1;
    m2 = (method**)v2;
    return method_get_calls (*m2) - method_get_calls (*m1);
}

/** Compare methods based on time used in method called from this method. */
static int method_compr_hold_time (const void* v1, const void* v2) {
    method** m1;
    method** m2;
    long l;
    m1 = (method**)v1;
    m2 = (method**)v2;
    if ((l = ((*m1)->time_used.tv_sec_hold - (*m2)->time_used.tv_sec_hold)) < 0) {
	return 1;
    } else if (l > 0) {
	return 0;
    } else {
	return (*m2)->time_used.tv_usec_hold - (*m1)->time_used.tv_usec_hold;
    }
}

/** Compare methods based on time used in this method and method called from this method. */
static int method_compr_total_time (const void* v1, const void* v2) {
    method** m1;
    method** m2;
    long l, lusec1, lsec1, lusec2, lsec2;
    m1 = (method**)v1;
    m2 = (method**)v2;

    lsec1 = (*m1)->time_used.tv_sec + (*m1)->time_used.tv_sec_hold;
    lusec1 = (*m1)->time_used.tv_usec + (*m1)->time_used.tv_usec_hold;
    if (lusec1 > 1000000) {
	lsec1++;
	lusec1 -= 1000000;
    }
    lsec2 = (*m2)->time_used.tv_sec + (*m2)->time_used.tv_sec_hold;
    lusec2 = (*m2)->time_used.tv_usec + (*m2)->time_used.tv_usec_hold;
    if (lusec2 > 1000000) {
	lsec2++;
	lusec2 -= 1000000;
    }    

    if ((l = (lsec1 - lsec2)) < 0) {
	return 1;
    } else if (l > 0) {
	return 0;
    } else {
	return lusec2 - lusec1;
    }
}

/** array of method compare functions... */
static int ((*method_comprs[])(const void* v1, const void* v2)) = { method_compr_class, 
								    method_compr_class, 
								    method_compr_time, 
								    method_compr_calls,
								    method_compr_hold_time,
								    method_compr_total_time };


/** The current method comparator */
static int (*method_compr) (const void* v1, const void* v2) = method_compr_time;


/** thread comparators */
static int jthread_compr_name (const void* v1, const void* v2) {
    jthread **t1 = (jthread**) v1;
    jthread **t2 = (jthread**) v2;
    return strcmp (jthread_get_thread_name(*t1), jthread_get_thread_name(*t2));
}

static int ((*jthread_comprs[])(const void* v1, const void* v2)) = { jthread_compr_name };
static int (*jthread_compr) (const void* v1, const void* v2) = jthread_compr_name;

static void set_jthread_compr (int i) {
    jthread_compr = jthread_comprs[i];
}

/** Count the number of classes that have more than zero instances. */
static void count_classes (void* data, cls_usage* cu) {
    cls* c;
    c = (cls*)data;
    if (cls_get_instances (c) != 0 && 
	filter_class_name(cls_get_name(c))) {
	classes_count++;
	cu->totalAlloced += c->usage.totalAlloced;
	cu->totalInstances += c->usage.totalInstances;
	cu->maximumInstances += c->usage.maximumInstances;
    }
}

/** Count the number of threads that are currently active */
static void count_threads (void* data) {
    jthread* t;
    t = (jthread*)data;
    /* do some check here */
    threads_count++;
}

/** Count the number of methodods that have used any time. */
static void count_methods (void* data) {
    method* m = (method*)data;
    if (!methodtime_is_zero (&m->time_used) &&
	filter_class_name(cls_get_name(method_get_owner(m)))) 
	methods_count++;
}

/** Add a row to the array of data. */
static void add_class_row (void* data) {
    cls* c;
    c = (cls*)data;    
    if (cls_get_instances (c) != 0 && 
	(c == total || filter_class_name(cls_get_name(c)))) 
	classlist[classes_count++] = c;
}

/** Add a row to the array of data. */
static void add_methods_row (void* data) {
    method* m = (method*)data;
    if (!methodtime_is_zero (&m->time_used) && 
	filter_class_name(cls_get_name(method_get_owner(m)))) 
	methodlist[methods_count++] = m;
}

/* Add a thread to the array of threads. */
static void add_threads_row (void* data) {
    jthread* t;
    t = (jthread*)data;
    /* do tests ? */
    threadslist[threads_count++] = t;
}

/** Add a row to the gtk CList (update the ui). */
static void add_row (cls* c, int row) {
    gchar *text[4]; 
    char cnt[64];
    char mcnt[64];
    char sze[64];
    int j;

    text[0] = cls_get_name (c);
    snprintf (cnt, 64, "%ld", cls_get_instances (c));
    text[1] = cnt;
    snprintf (mcnt, 64, "%ld", cls_get_max_instances (c));
    text[2] = mcnt;
    snprintf (sze, 64, "%ld", cls_get_size (c));
    text[3] = sze;

    if (row < object_list_size) {
	j = 0;
	if (gtk_clist_get_row_data (GTK_CLIST (object_list), row) == c)
	    j++;
	for (; j < 4; j++)
	    gtk_clist_set_text (GTK_CLIST (object_list), row, j, text[j]);
    } else {
	object_list_size = gtk_clist_append ((GtkCList*)object_list, text) + 1;
    }

    gtk_clist_set_row_data (GTK_CLIST (object_list), row, c);
}

/** Dump a class to the given file. */
static void dump_row (cls* c, FILE* f) {
    fprintf (f, "%s\t%ld\t%ld\t%ld\n", 
	     cls_get_name (c),
	     cls_get_instances (c),
	     cls_get_max_instances (c),
	     cls_get_size (c));    
}

/** Add a row to the gtk CList (update the ui). */
static void add_method_row (method* m, int row) {
    cls* c;
    gchar *text[6]; 
    char sec[64];
    char calls[64];
    char hsec[64];
    char tsec[64];
    long lsec;
    long lusec;
    methodtime* mt;
    int i = 0;
    int j;

    c = method_get_owner (m);
    text[i++] = cls_get_name (c);
    text[i++] = method_get_method_jmpname (m);
    mt = method_get_time_used (m);
    snprintf (sec, 64, "%ld.%06ld", mt->tv_sec, mt->tv_usec);
    text[i++] = sec;

    snprintf (calls, 64, "%ld", method_get_calls (m));
    text[i++] = calls;

    snprintf (hsec, 64, "%ld.%06ld", mt->tv_sec_hold, mt->tv_usec_hold);
    text[i++] = hsec;

    lusec = mt->tv_usec_hold + mt->tv_usec;
    lsec = mt->tv_sec_hold + mt->tv_sec;
    if (lusec > 1000000) {
	lsec++;
	lusec -= 1000000;
    }
    snprintf (tsec, 64, "%ld.%06ld", lsec, lusec);
    text[i++] = tsec;
    
    if (row < mlist_size) {
	j = 0;
	if (gtk_clist_get_row_data (GTK_CLIST (mlist), row) == m)
	    j = 2;
	for (; j < 6; j++)
	    gtk_clist_set_text (GTK_CLIST (mlist), row, j, text[j]);
    } else {
	mlist_size = gtk_clist_append ((GtkCList*)mlist, text) + 1;
    }
    
    gtk_clist_set_row_data (GTK_CLIST (mlist), row, m);
}

/** dump a method to the given file */
static void dump_method_row (method* m, FILE* f) {
    methodtime* mt;
    cls* c = method_get_owner (m);
    mt = method_get_time_used (m);
    fprintf (f, "%s\t%s\t%ld.%06ld\t%ld\t%ld.%06ld\n", 
	     cls_get_name (c),
	     method_get_method_jmpname (m),
	     mt->tv_sec,
	     mt->tv_usec,
	     method_get_calls (m),
	     mt->tv_sec_hold,
	     mt->tv_usec_hold);	     
}

static void dump_jthread_row (jthread* t, FILE* f) {
    method* mp = NULL;
    timerstack* s;
    int i;

    fprintf (f, "%p\t%s\t%s\t%s\t\n",
	     t->env_id,
	     jthread_get_thread_name(t),
	     jthread_get_group_name(t),
	     jthread_get_parent_name(t));
    
    s = get_timerstack ((JNIEnv*)t->env_id);
    for (i = s->top - 1; i >= 0; i--) {
	methodtime* mt = s->times + i;
	mp = get_method (mt->method_id);
	if (mp != NULL) 
	    fprintf (f, "\t%s\t%s\n",
		     cls_get_name (mp->owner),
		     method_get_method_jmpname (mp));
	else 
	    fprintf (f, "\t?\t?\n" );	    	
    }
}


void set_jvmpi_status (char *text) {
    if (statusbar) {
	gdk_threads_enter ();
	set_status(text); 
	gdk_threads_leave ();
    }
}

/** Set the status bar text. */
void set_status (char* text) {
    if (statusbar) {
	guint context_id = 1;        
	gtk_statusbar_pop (GTK_STATUSBAR (statusbar), context_id);
	gtk_statusbar_push (GTK_STATUSBAR (statusbar), context_id, text);
    }
}

/** Update the tables. */
void updateUI (hashtab* classes, hashtab* methods) {
    int i;
    int oldcount = classes_count;    
    
    if (!init_done)
	return;

    if (!is_updating)
	return;

    gdk_threads_enter ();

    if (tracing_objects ()) {
	cls_usage cu;
	/* update the object list window.. */
	gtk_clist_freeze ((GtkCList*)object_list);
	classes_count = 0;
	cu.totalAlloced = 0;
	cu.totalInstances = 0;
	cu.maximumInstances = 0;
	hash_for_each_with_arg ((hash_iter_fa)count_classes, classes, &cu); 
	classes_count++;
	total->usage = cu;
	if (oldcount != classes_count) 
	    classlist = realloc (classlist, classes_count * sizeof (cls*));
	classes_count = 0;
	add_class_row (total);
	hash_for_each ((hash_iter_f)add_class_row, classes); 
    
	qsort (classlist, classes_count, sizeof (cls*), cls_compr);
	
	for (i = 0; i < classes_count; i++) 
	    add_row (classlist[i], i);
	if (classes_count < object_list_size) {
	    for (i = object_list_size - 1; i >= classes_count; i--)
		gtk_clist_remove (GTK_CLIST (object_list), i);
	    object_list_size=classes_count;
	}
	gtk_clist_thaw ((GtkCList*)object_list);
    }

    if (tracing_methods ()) {
	/* update the method list window.. */
	oldcount = methods_count;
	gtk_clist_freeze ((GtkCList*)mlist);    
	oldcount = methods_count;    
	methods_count = 0;

	hash_for_each ((hash_iter_f)count_methods, methods);
	if (oldcount != methods_count) {
	    methodlist = realloc (methodlist, methods_count * sizeof (method*));
	}
	methods_count = 0;
	hash_for_each ((hash_iter_f)add_methods_row, methods);
	qsort (methodlist, methods_count, sizeof (method*), method_compr);
	for (i = 0; i < methods_count; i++) 
	    add_method_row (methodlist[i], i);
	if (methods_count < mlist_size) {
	    for (i = mlist_size - 1; i >= methods_count; i--)
		gtk_clist_remove (GTK_CLIST (mlist), i);
	    mlist_size = methods_count;
	}
	gtk_clist_thaw ((GtkCList*)mlist);
    }

    if (threadswin != NULL)
	update_threads_window (threadswin, get_threads ());
    
    gdk_threads_leave ();
}

int dump_data (char* dumpfile, hashtab* classes, hashtab* methods, hashtab* threads) {
    FILE* f;
    int i;
    int oldcount = classes_count;
    char filename[128];
    char status[128];
    cls_usage cu;
    snprintf (filename, 128, "%s-%d.txt", 
	      (dumpfile == NULL ? "jmp_dump" : dumpfile), ++dump_id);    
    snprintf (status, 128, _("dumped to: %s"), filename);
    set_status (status);
    f = fopen (filename, "w");
    if (f == NULL)
	return -1;

    fprintf (f, "Threads\n");
    fprintf (f, "thread\tgroup\tparent\tclass\tmethod\n");
    fprintf (f, "--------------------------------------------------------------\n");
    oldcount = threads_count;
    threads_count = 0;
    hash_for_each ((hash_iter_f)count_threads, threads);
    if (oldcount != threads_count) 
	threadslist = realloc (threadslist, threads_count * sizeof(jthread*));
    threads_count = 0;
    hash_for_each ((hash_iter_f)add_threads_row, threads);
    qsort(threadslist, threads_count, sizeof(jthread*), jthread_compr);
    for (i = 0; i < threads_count; i++){
	dump_jthread_row (threadslist[i], f);
    }

    fprintf (f, "\n\n\nClassdump\n");
    fprintf (f, "class_name\t#instaces\tmax #instances\tsize\n");
    fprintf (f, "--------------------------------------------------------------\n");
    hash_for_each_with_arg ((hash_iter_fa)count_classes, classes, &cu); 
    if (oldcount != classes_count) 
	classlist = realloc (classlist, classes_count * sizeof (cls*));
    classes_count = 0;
    hash_for_each ((hash_iter_f)add_class_row, classes); 
    qsort (classlist, classes_count, sizeof (cls*), cls_compr);
    for (i = 0; i < classes_count; i++) 
	dump_row (classlist[i], f);

    fprintf (f, "\n\n\nMethods\n");
    fprintf (f, "class name\tmethod\tsec\tcalls\tsubsec\n");
    fprintf (f, "--------------------------------------------------------------\n");
    oldcount = methods_count;    
    methods_count = 0;
    hash_for_each ((hash_iter_f)count_methods, methods);
    if (oldcount != methods_count) 
	methodlist = realloc (methodlist, methods_count * sizeof (method*));
    methods_count = 0;
    hash_for_each ((hash_iter_f)add_methods_row, methods);
    qsort (methodlist, methods_count, sizeof (method*), method_compr);
    for (i = 0; i < methods_count; i++) 
	dump_method_row (methodlist[i], f);

    fflush (f);
    fclose (f);
    return dump_id;
}

void set_dump_timer (int dt) {
    dumptimer = dt;
}

int get_dump_timer () {
    return dumptimer;
}

static gint delete_event (GtkWidget *widget, GdkEvent  *event, gpointer   data) {
    return(TRUE);
}

static void destroy (GtkWidget *widget, gpointer data) {
    gtk_main_quit();
}

static void cls_column_clicked (GtkWidget *widget, gint column) {
    if (cls_comprs[column] != NULL)
	cls_compr = cls_comprs[column];
    else 
	fprintf (stdout, "Sort order not yet implemented.\n");
}

static void method_column_clicked (GtkWidget *widget, gint column) {
    if (method_comprs[column] != NULL)
	method_compr = method_comprs[column];
    else 
	fprintf (stdout, "Sort order not yet implemented.\n");
}


void run_dump (GtkMenuItem *menuitem, gpointer user_data) {
    run_data_dump ();
}

void run_garbage_collector (GtkMenuItem *menuitem, gpointer user_data) {
    set_status (_("Requesting garbage collection ..."));
    run_GC ();
}

void run_heap_dumper (GtkMenuItem *menuitem, gpointer user_data) {
    set_status (_("Getting heap dump"));
    run_heap_dump ();
    set_status (_("Heap dumped"));
}

void run_reset_counter (GtkMenuItem *menuitem, gpointer user_data) {
    set_status (_("Resetting counters..."));
    reset_counters ();
    set_status (_("Counters reseted"));
}

void run_restore_counter (GtkMenuItem *menuitem, gpointer user_data) {
    set_status (_("Resetoring counters..."));
    restore_counters ();
    set_status (_("Counters restored"));
}

static void freeze_ui () {
    is_updating = !is_updating;
    set_status (is_updating ? _("UI updating continued") : _("UI updating frozen"));
}

static void show_threads () {
    set_status ("showing threads...");
    if (threadswin == NULL)
	threadswin = create_threads_window (get_threads ());
    else 
	update_threads_window (threadswin, get_threads ());
    gtk_widget_show (GTK_WIDGET (threadswin->window));
}

/** Originally create by glade, which seems like a nice tool... 
 */
GtkWidget* create_about_dialog (void) {
    GtkWidget *about_dialog;
    GtkWidget *dialog_vbox1;
    GtkWidget *label1;
    GtkWidget *dialog_action_area1;
    GtkWidget *hbuttonbox2;
    guint button6_key;
    GtkWidget *button6;
    GtkAccelGroup *accel_group;

    accel_group = gtk_accel_group_new ();

    about_dialog = gtk_dialog_new ();
    gtk_object_set_data (GTK_OBJECT (about_dialog), "about_dialog", about_dialog);
    gtk_window_set_title (GTK_WINDOW (about_dialog), _("About JMP"));
    gtk_window_set_policy (GTK_WINDOW (about_dialog), TRUE, TRUE, FALSE);

    dialog_vbox1 = GTK_DIALOG (about_dialog)->vbox;
    gtk_object_set_data (GTK_OBJECT (about_dialog), "dialog_vbox1", dialog_vbox1);
    gtk_widget_show (dialog_vbox1);

    label1 = gtk_label_new (_("\nJMP - A java profiler by\n  Robert 'Ernimril' Olofsson - d94-rol@nada.kth.se  \n     Bodo Pfelzer \n"));
    gtk_widget_ref (label1);
    gtk_object_set_data_full (GTK_OBJECT (about_dialog), "label1", label1,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (label1);
    gtk_box_pack_start (GTK_BOX (dialog_vbox1), label1, FALSE, FALSE, 0);

    dialog_action_area1 = GTK_DIALOG (about_dialog)->action_area;
    gtk_object_set_data (GTK_OBJECT (about_dialog), "dialog_action_area1", dialog_action_area1);
    gtk_widget_show (dialog_action_area1);
    gtk_container_set_border_width (GTK_CONTAINER (dialog_action_area1), 10);

    hbuttonbox2 = gtk_hbutton_box_new ();
    gtk_widget_ref (hbuttonbox2);
    gtk_object_set_data_full (GTK_OBJECT (about_dialog), "hbuttonbox2", hbuttonbox2,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (hbuttonbox2);
    gtk_box_pack_start (GTK_BOX (dialog_action_area1), hbuttonbox2, TRUE, TRUE, 0);

    button6 = gtk_button_new_with_label ("");
    button6_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (button6)->child),
					 _("_Ok"));
    gtk_widget_add_accelerator (button6, "clicked", accel_group,
				button6_key, GDK_MOD1_MASK, 0);
    gtk_widget_ref (button6);
    gtk_object_set_data_full (GTK_OBJECT (about_dialog), "button6", button6,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (button6);
    gtk_container_add (GTK_CONTAINER (hbuttonbox2), button6);
    GTK_WIDGET_SET_FLAGS (button6, GTK_CAN_DEFAULT);
    gtk_widget_add_accelerator (button6, "clicked", accel_group,
				GDK_Return, 0,
				GTK_ACCEL_VISIBLE);
    gtk_widget_add_accelerator (button6, "clicked", accel_group,
				GDK_space, 0,
				GTK_ACCEL_VISIBLE);

    gtk_signal_connect_object (GTK_OBJECT (button6), "clicked",
			       GTK_SIGNAL_FUNC (gtk_widget_hide),
			       GTK_OBJECT (about_dialog));

    gtk_window_add_accel_group (GTK_WINDOW (about_dialog), accel_group);

    return about_dialog;
}

static void about_jmp (GtkMenuItem *menuitem, gpointer user_data) {
    if (about == NULL)
	about = create_about_dialog ();	
    gtk_widget_show (about);
}

/** Originally create by glade, which seems like a nice tool... 
 *  Names of widgets is not totally nice though...
 */
GtkWidget* create_JMP (void)
{
    GtkWidget *JMP;
    GtkWidget *alignment1;
    GtkWidget *vbox1;
    GtkWidget *menubar1;
    guint tmp_key;
    GtkWidget *file1;
    GtkMenuItem *options1;
    GtkMenu *options1_menu;
    GtkMenuItem *options1_subitem;
    GtkWidget *file1_menu;
    GtkAccelGroup *file1_menu_accels;
    GtkWidget *dump1;
    GtkWidget *run_garbage_collector1;
    GtkWidget *run_freeze_ui;
    GtkWidget *run_show_threads;
    GtkWidget *run_reset_counters;
    GtkWidget *run_restore_counters;
    GtkWidget *help1;
    GtkWidget *help1_menu;
    GtkAccelGroup *help1_menu_accels;
    GtkWidget *heap_menu;
    GtkWidget *about_jmp1;
    GtkWidget *hbuttonbox1;
    guint button3_key;
    GtkWidget *button3;
    guint buttonR_key;
    GtkWidget *buttonR;
    guint buttonS_key;
    GtkWidget *buttonS;
    guint button4_key;
    GtkWidget *button4;
    guint button_heap_key;
    GtkWidget *button_heap;
    guint button5_key;
    GtkWidget *button5;
    guint button6_key;
    GtkWidget *button6;
    guint button7_key;
    GtkWidget *button7;
    GtkAccelGroup *accel_group;

    accel_group = gtk_accel_group_new ();

    JMP = gtk_window_new (GTK_WINDOW_TOPLEVEL);
    gtk_object_set_data (GTK_OBJECT (JMP), "JMP", JMP);
    gtk_window_set_title (GTK_WINDOW (JMP), _("Java Memory Profiler - Main"));

    alignment1 = gtk_alignment_new (0.5, 0.5, 1, 1);
    gtk_widget_ref (alignment1);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "alignment1", alignment1,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (alignment1);
    gtk_container_add (GTK_CONTAINER (JMP), alignment1);

    vbox1 = gtk_vbox_new (FALSE, 0);
    gtk_widget_ref (vbox1);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "vbox1", vbox1,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (vbox1);
    gtk_container_add (GTK_CONTAINER (alignment1), vbox1);

    menubar1 = gtk_menu_bar_new ();
    gtk_widget_ref (menubar1);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "menubar1", menubar1,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (menubar1);
    gtk_box_pack_start (GTK_BOX (vbox1), menubar1, FALSE, FALSE, 0);

    file1 = gtk_menu_item_new_with_label ("");
    tmp_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (file1)->child),
				     _("_File"));
    gtk_widget_add_accelerator (file1, "activate_item", accel_group,
				tmp_key, GDK_MOD1_MASK, 0);
    gtk_widget_ref (file1);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "file1", file1,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (file1);
    gtk_container_add (GTK_CONTAINER (menubar1), file1);

    file1_menu = gtk_menu_new ();
    gtk_widget_ref (file1_menu);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "file1_menu", file1_menu,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_menu_item_set_submenu (GTK_MENU_ITEM (file1), file1_menu);
    file1_menu_accels = gtk_menu_ensure_uline_accel_group (GTK_MENU (file1_menu));

    dump1 = gtk_menu_item_new_with_label ("");
    tmp_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (dump1)->child),
				     _("_Dump"));
    gtk_widget_add_accelerator (dump1, "activate_item", file1_menu_accels,
				tmp_key, 0, 0);
    gtk_widget_ref (dump1);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "dump1", dump1,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (dump1);
    gtk_container_add (GTK_CONTAINER (file1_menu), dump1);

    run_reset_counters = gtk_menu_item_new_with_label ("");
    tmp_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (run_reset_counters)->child),
				     _("_Reset counters"));
    gtk_widget_add_accelerator (run_reset_counters, "activate_item", file1_menu_accels,
				tmp_key, 0, 0);
    gtk_widget_ref (run_reset_counters);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "run_reset_counters", run_reset_counters,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (run_reset_counters);
    gtk_container_add (GTK_CONTAINER (file1_menu), run_reset_counters);

    run_restore_counters = gtk_menu_item_new_with_label ("");
    tmp_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (run_restore_counters)->child),
				     _("Re_store counters"));
    gtk_widget_add_accelerator (run_restore_counters, "activate_item", file1_menu_accels,
				tmp_key, 0, 0);
    gtk_widget_ref (run_restore_counters);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "run_restore_counters", run_restore_counters,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (run_restore_counters);
    gtk_container_add (GTK_CONTAINER (file1_menu), run_restore_counters);

    run_garbage_collector1 = gtk_menu_item_new_with_label ("");
    tmp_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (run_garbage_collector1)->child),
				     _("System._GC"));
    gtk_widget_add_accelerator (run_garbage_collector1, "activate_item", file1_menu_accels,
				tmp_key, 0, 0);
    gtk_widget_ref (run_garbage_collector1);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "run_garbage_collector1", run_garbage_collector1,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (run_garbage_collector1);
    gtk_container_add (GTK_CONTAINER (file1_menu), run_garbage_collector1);

    heap_menu = gtk_menu_item_new_with_label ("");
    tmp_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (heap_menu)->child),
				     _("_Heapdump"));
    gtk_widget_add_accelerator (heap_menu, "activate_item", file1_menu_accels,
				tmp_key, 0, 0);
    gtk_widget_ref (heap_menu);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "heap_menu", heap_menu,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (heap_menu);
    gtk_container_add (GTK_CONTAINER (file1_menu), heap_menu);

    run_freeze_ui = gtk_menu_item_new_with_label ("");
    tmp_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (run_freeze_ui)->child),
				     _("_Freeze UI"));
    gtk_widget_add_accelerator (run_freeze_ui, "activate_item", file1_menu_accels,
				tmp_key, 0, 0);
    gtk_widget_ref (run_freeze_ui);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "run_freeze_ui", run_freeze_ui, 
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (run_freeze_ui);
    gtk_container_add (GTK_CONTAINER (file1_menu), run_freeze_ui);

    run_show_threads = gtk_menu_item_new_with_label ("");
    tmp_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (run_show_threads)->child),
				     _("Threads"));
    gtk_widget_add_accelerator (run_show_threads, "activate_item", file1_menu_accels,
				tmp_key, 0, 0);
    gtk_widget_ref (run_show_threads);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "run_show_threads", run_show_threads, 
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (run_show_threads);
    gtk_container_add (GTK_CONTAINER (file1_menu), run_show_threads);

    /* Add the options menu */
    options1_menu = GTK_MENU(gtk_menu_new());
    options1 = GTK_MENU_ITEM(gtk_menu_item_new_with_label(""));
    tmp_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN(options1)->child),
				     _("_Options"));
    gtk_widget_add_accelerator (GTK_WIDGET (options1), "activate_item",
			       accel_group, tmp_key, GDK_MOD1_MASK, 0);
    gtk_menu_item_set_submenu (options1, GTK_WIDGET (options1_menu));
    gtk_widget_show (GTK_WIDGET (options1));
    options1_subitem = GTK_MENU_ITEM (gtk_menu_item_new_with_label (_("Filter")));
    gtk_widget_show (GTK_WIDGET (options1_subitem));
    gtk_signal_connect_object (GTK_OBJECT (options1_subitem),"activate",
			       GTK_SIGNAL_FUNC (filter_edit_options), NULL);
    gtk_menu_append (options1_menu, GTK_WIDGET (options1_subitem));

    gtk_menu_bar_append (GTK_MENU_BAR (menubar1), GTK_WIDGET (options1));

    /* Add the help menu... */
    help1 = gtk_menu_item_new_with_label ("");
    tmp_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (help1)->child),
				     _("_Help"));
    gtk_widget_add_accelerator (help1, "activate_item", accel_group,
				tmp_key, GDK_MOD1_MASK, 0);
    gtk_widget_ref (help1);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "help1", help1,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (help1);
    gtk_container_add (GTK_CONTAINER (menubar1), help1);

    help1_menu = gtk_menu_new ();
    gtk_widget_ref (help1_menu);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "help1_menu", help1_menu,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_menu_item_set_submenu (GTK_MENU_ITEM (help1), help1_menu);
    help1_menu_accels = gtk_menu_ensure_uline_accel_group (GTK_MENU (help1_menu));

    about_jmp1 = gtk_menu_item_new_with_label ("");
    tmp_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (about_jmp1)->child),
				     _("_About JMP"));
    gtk_widget_add_accelerator (about_jmp1, "activate_item", help1_menu_accels,
				tmp_key, 0, 0);
    gtk_widget_ref (about_jmp1);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "about_jmp1", about_jmp1,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (about_jmp1);
    gtk_container_add (GTK_CONTAINER (help1_menu), about_jmp1);

    hbuttonbox1 = gtk_hbutton_box_new ();
    gtk_hbutton_box_set_spacing_default (0);
    gtk_widget_ref (hbuttonbox1);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "hbuttonbox1", hbuttonbox1,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (hbuttonbox1);
    gtk_box_pack_start (GTK_BOX (vbox1), hbuttonbox1, TRUE, TRUE, 0);

    button3 = gtk_button_new_with_label ("");
    button3_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (button3)->child),
					 _("_Dump"));
    gtk_widget_add_accelerator (button3, "clicked", accel_group,
				button3_key, GDK_MOD1_MASK, 0);
    gtk_widget_ref (button3);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "button3", button3,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_tooltips_set_tip(tooltips, button3, _("Dump data to a text file"), NULL);
    gtk_widget_show (button3);
    gtk_container_add (GTK_CONTAINER (hbuttonbox1), button3);
    GTK_WIDGET_SET_FLAGS (button3, GTK_CAN_DEFAULT);

    buttonR = gtk_button_new_with_label ("");
    buttonR_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (buttonR)->child),
					 _("_Reset"));
    gtk_widget_add_accelerator (buttonR, "clicked", accel_group,
				buttonR_key, GDK_MOD1_MASK, 0);
    gtk_widget_ref (buttonR);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "buttonR", buttonR,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_tooltips_set_tip(tooltips, buttonR, _("Reset counters to zero"), NULL);
    gtk_widget_show (buttonR);
    gtk_container_add (GTK_CONTAINER (hbuttonbox1), buttonR);
    GTK_WIDGET_SET_FLAGS (buttonR, GTK_CAN_DEFAULT);

    buttonS = gtk_button_new_with_label ("");
    buttonS_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (buttonS)->child),
					 _("Re_store"));
    gtk_widget_add_accelerator (buttonS, "clicked", accel_group,
				buttonS_key, GDK_MOD1_MASK, 0);
    gtk_widget_ref (buttonS);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "buttonS", buttonS,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_tooltips_set_tip(tooltips, buttonS, _("Restore counters to the standard (full) values"), NULL);
    gtk_widget_show (buttonS);
    gtk_container_add (GTK_CONTAINER (hbuttonbox1), buttonS);
    GTK_WIDGET_SET_FLAGS (buttonS, GTK_CAN_DEFAULT);

    button4 = gtk_button_new_with_label ("");
    button4_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (button4)->child),
					 _("System._GC"));
    gtk_widget_add_accelerator (button4, "clicked", accel_group,
				button4_key, GDK_MOD1_MASK, 0);
    gtk_widget_ref (button4);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "button4", button4,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_tooltips_set_tip(tooltips, button4, _("Run the java virtual machines garbage collector"), NULL);
    gtk_widget_show (button4);
    gtk_container_add (GTK_CONTAINER (hbuttonbox1), button4);
    GTK_WIDGET_SET_FLAGS (button4, GTK_CAN_DEFAULT);

    button_heap = gtk_button_new_with_label ("");
    button_heap_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (button_heap)->child),
					     _("_Heapdump"));
    gtk_widget_add_accelerator (button_heap, "clicked", accel_group,
				button_heap_key, GDK_MOD1_MASK, 0);
    gtk_widget_ref (button_heap);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "button_heap", button_heap,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_tooltips_set_tip(tooltips, button_heap, _("Show the current heap"), NULL);
    gtk_widget_show (button_heap);
    gtk_container_add (GTK_CONTAINER (hbuttonbox1), button_heap);
    GTK_WIDGET_SET_FLAGS (button_heap, GTK_CAN_DEFAULT);

    button5 = gtk_button_new_with_label ("");
    button5_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (button5)->child),
					 _("_About"));
    gtk_widget_add_accelerator (button5, "clicked", accel_group,
				button5_key, GDK_MOD1_MASK, 0);
    gtk_widget_ref (button5);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "button5", button5,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_tooltips_set_tip(tooltips, button5, _("About JMP"), NULL);
    gtk_widget_show (button5);
    gtk_container_add (GTK_CONTAINER (hbuttonbox1), button5);
    GTK_WIDGET_SET_FLAGS (button5, GTK_CAN_DEFAULT);

    button6 = gtk_button_new_with_label ("");
    button6_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (button6)->child),
					 _("_Freeze ui"));
    gtk_widget_add_accelerator (button6, "clicked", accel_group,
				button6_key, GDK_MOD1_MASK, 0);
    gtk_widget_ref (button6);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "button6", button6,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_tooltips_set_tip(tooltips, button6, _("Stop updating values in the JMP windows"), NULL);
    gtk_widget_show (button6);
    gtk_container_add (GTK_CONTAINER (hbuttonbox1), button6);
    GTK_WIDGET_SET_FLAGS (button6, GTK_CAN_DEFAULT);

    button7 = gtk_button_new_with_label ("");
    button7_key = gtk_label_parse_uline (GTK_LABEL (GTK_BIN (button7)->child),
					 _("Threads"));
    gtk_widget_add_accelerator (button7, "clicked", accel_group,
				button7_key, GDK_MOD1_MASK, 0);
    gtk_widget_ref (button7);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "button7", button7,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_tooltips_set_tip(tooltips, button7, _("Show the current threads"), NULL);
    gtk_widget_show (button7);
    gtk_container_add (GTK_CONTAINER (hbuttonbox1), button7);
    GTK_WIDGET_SET_FLAGS (button7, GTK_CAN_DEFAULT);


    statusbar = gtk_statusbar_new ();
    gtk_widget_ref (statusbar);
    gtk_object_set_data_full (GTK_OBJECT (JMP), "statusbar", statusbar,
			      (GtkDestroyNotify) gtk_widget_unref);
    gtk_widget_show (statusbar);
    gtk_box_pack_start (GTK_BOX (vbox1), statusbar, FALSE, FALSE, 0);

    gtk_signal_connect (GTK_OBJECT (dump1), "activate",
			GTK_SIGNAL_FUNC (run_dump),
			NULL);
    gtk_signal_connect (GTK_OBJECT (run_garbage_collector1), "activate",
			GTK_SIGNAL_FUNC (run_garbage_collector),
			NULL);
    gtk_signal_connect (GTK_OBJECT (heap_menu), "activate",
			GTK_SIGNAL_FUNC (run_heap_dumper),
			NULL);
    gtk_signal_connect (GTK_OBJECT (run_freeze_ui), "activate",
			GTK_SIGNAL_FUNC (freeze_ui),
			NULL);
    gtk_signal_connect (GTK_OBJECT (run_show_threads), "activate",
			GTK_SIGNAL_FUNC (show_threads),
			NULL);
    gtk_signal_connect (GTK_OBJECT (run_reset_counters), "activate",
			GTK_SIGNAL_FUNC (run_reset_counter),
			NULL);
    gtk_signal_connect (GTK_OBJECT (run_restore_counters), "activate",
			GTK_SIGNAL_FUNC (run_restore_counter),
			NULL);
    gtk_signal_connect (GTK_OBJECT (about_jmp1), "activate",
			GTK_SIGNAL_FUNC (about_jmp),
			NULL);
    gtk_signal_connect (GTK_OBJECT (button3), "clicked",
			GTK_SIGNAL_FUNC (run_dump),
			NULL);
    gtk_signal_connect (GTK_OBJECT (buttonR), "clicked",
			GTK_SIGNAL_FUNC (run_reset_counter),
			NULL);
    gtk_signal_connect (GTK_OBJECT (buttonS), "clicked",
			GTK_SIGNAL_FUNC (run_restore_counter),
			NULL);
    gtk_signal_connect (GTK_OBJECT (button4), "clicked",
			GTK_SIGNAL_FUNC (run_garbage_collector),
			NULL);
    gtk_signal_connect (GTK_OBJECT (button_heap), "clicked",
			GTK_SIGNAL_FUNC (run_heap_dumper),
			NULL);
    gtk_signal_connect (GTK_OBJECT (button5), "clicked",
			GTK_SIGNAL_FUNC (about_jmp),
			NULL);
    gtk_signal_connect (GTK_OBJECT (button6), "clicked",
			GTK_SIGNAL_FUNC (freeze_ui),
			NULL);
    gtk_signal_connect (GTK_OBJECT (button7), "clicked",
			GTK_SIGNAL_FUNC (show_threads),
			NULL);

    gtk_window_add_accel_group (GTK_WINDOW (JMP), accel_group);

    return JMP;
}

/** This is the timeout function called from gtk-main.
 *  Check if it is time to quit.
 */
gint run_updates (gpointer data) {
    if (quit) {
	gtk_main_quit ();   /* time to shut down. */
	return FALSE;       /* stop timer. */
    }

    if (usingUI ())
	updateUI (get_classes (), get_methods ());

    if (dumptimer > 0) {
	dumptime++;
	if (dumptime == dumptimer) {
	    run_data_dump ();
	    dumptime = 0;
	} 
    }

    return TRUE;            /* keep timer */
}

/** Update state. 
 */
void quit_ui (void) {
    quit = 1;
    if (tracing_objects ())
	cls_free (total);
}

/** Start the UI thread up. Builds windows that should be visible.
 *  Enter gtk_main ()
 */
void gtkthread (void* data) {
    int argc = 1;
    char** argv;
    gchar *cls_titles[] = { _("Class"), _("Instances"), _("Max instances"), _("Size")};
    gchar *titles[] = { _("Class"), _("Method"), _("secs"), _("calls"), _("subs sec"), _("total")};
    argv = malloc (sizeof (*argv));
    argv[0] = "jmp";

    g_thread_init (NULL);
    gtk_init (&argc, &argv);

    if (tracing_objects () && usingUI ()) {
	total = cls_new (_("Total"), _("Total"), NULL, 0, 0, NULL, 0, NULL);
	jmpwin = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_signal_connect (GTK_OBJECT (jmpwin), "delete_event",
			    GTK_SIGNAL_FUNC (delete_event), NULL);
	gtk_signal_connect (GTK_OBJECT (jmpwin), "destroy",
			    GTK_SIGNAL_FUNC (destroy), NULL);

	gtk_object_set_data (GTK_OBJECT (jmpwin), "jmpwin", jmpwin);
	gtk_window_set_title (GTK_WINDOW (jmpwin), _("Java Memory Profiler - Objects"));
    
	scroller = gtk_scrolled_window_new (NULL, NULL);
	gtk_widget_ref (scroller);
	gtk_object_set_data_full (GTK_OBJECT (jmpwin), "scroller", scroller,
				  (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (scroller);
	gtk_container_add (GTK_CONTAINER (jmpwin), scroller);
    
	object_list = gtk_clist_new_with_titles (4, cls_titles);
	gtk_object_set_data_full (GTK_OBJECT (jmpwin), "object_list", object_list,
				  (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (object_list);

	gtk_clist_set_column_justification((GtkCList*)object_list, 1, GTK_JUSTIFY_RIGHT);
	gtk_clist_set_column_justification((GtkCList*)object_list, 2, GTK_JUSTIFY_RIGHT);
	gtk_clist_set_column_justification((GtkCList*)object_list, 3, GTK_JUSTIFY_RIGHT);    

	gtk_clist_set_column_width (GTK_CLIST (object_list), 0, 200);
	gtk_clist_set_column_width (GTK_CLIST (object_list), 1, 80);
	gtk_clist_set_column_width (GTK_CLIST (object_list), 2, 80);
	gtk_clist_set_column_width (GTK_CLIST (object_list), 3, 80);
	
	gtk_container_add (GTK_CONTAINER (scroller), object_list);
	gtk_signal_connect (GTK_OBJECT (object_list), "click_column", 
			    GTK_SIGNAL_FUNC (cls_column_clicked),
			    NULL);    
	
	gtk_signal_connect (GTK_OBJECT (object_list),
			    "select_row",
			    GTK_SIGNAL_FUNC (olist_row_selected),
			    NULL);
	gtk_signal_connect (GTK_OBJECT (object_list),
			    "unselect_row",
			    GTK_SIGNAL_FUNC (olist_row_unselected),
			    NULL);
	gtk_signal_connect (GTK_OBJECT (object_list), "button_press_event",
			    GTK_SIGNAL_FUNC (olist_button_handler), 
			    GTK_WIDGET (object_list));

	gtk_widget_set_usize (jmpwin, 500, 200);
	gtk_widget_show (jmpwin);
    }    

    if (tracing_methods () && usingUI ()) {
	mwin = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	gtk_signal_connect (GTK_OBJECT (mwin), "delete_event",
			    GTK_SIGNAL_FUNC (delete_event), NULL);
	gtk_signal_connect (GTK_OBJECT (mwin), "destroy",
			    GTK_SIGNAL_FUNC (destroy), NULL);
	gtk_window_set_title (GTK_WINDOW (mwin), _("Java Memory Profiler - Methods"));
	mscroller = gtk_scrolled_window_new (NULL, NULL);
	gtk_object_set_data_full (GTK_OBJECT (mwin), "mscroller", mscroller,
				  (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (mscroller);
	gtk_container_add (GTK_CONTAINER (mwin), mscroller);
	mlist = gtk_clist_new_with_titles (6, titles);
	gtk_object_set_data_full (GTK_OBJECT (mwin), "mlist", mlist,
				  (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show (mlist);

	gtk_clist_set_column_width (GTK_CLIST (mlist), 0, 200);
	gtk_clist_set_column_width (GTK_CLIST (mlist), 1, 200);
	gtk_clist_set_column_width (GTK_CLIST (mlist), 2, 80);
	gtk_clist_set_column_width (GTK_CLIST (mlist), 3, 80);
	gtk_clist_set_column_width (GTK_CLIST (mlist), 4, 80);
	gtk_clist_set_column_width (GTK_CLIST (mlist), 5, 80);
	
	gtk_clist_set_column_justification((GtkCList*)mlist, 2, GTK_JUSTIFY_RIGHT);
	gtk_clist_set_column_justification((GtkCList*)mlist, 3, GTK_JUSTIFY_RIGHT);
	gtk_clist_set_column_justification((GtkCList*)mlist, 4, GTK_JUSTIFY_RIGHT);    
	gtk_clist_set_column_justification((GtkCList*)mlist, 5, GTK_JUSTIFY_RIGHT);    

	gtk_container_add (GTK_CONTAINER (mscroller), mlist);
	gtk_signal_connect (GTK_OBJECT (mlist), "click_column", 
			    GTK_SIGNAL_FUNC (method_column_clicked),
			    NULL);    

	gtk_signal_connect (GTK_OBJECT (mlist),
			    "select_row",
			    GTK_SIGNAL_FUNC (mlist_row_selected),
			    NULL);
	gtk_signal_connect (GTK_OBJECT (mlist),
			    "unselect_row",
			    GTK_SIGNAL_FUNC (mlist_row_unselected),
			    NULL);
	gtk_signal_connect (GTK_OBJECT(mlist), "button_press_event",
			    GTK_SIGNAL_FUNC (mlist_button_handler), 
			    NULL);

	gtk_widget_set_usize (mwin, 800, 200);
	gtk_widget_show (mwin);
    }
    
    if (usingUI ()) {
	tooltips = gtk_tooltips_new ();
	JMPmain = create_JMP ();
	gtk_widget_show (JMPmain);
	set_jthread_compr (0);
    }	

    gtk_timeout_add (1000, run_updates, NULL);
    init_done = 1;
    gdk_threads_enter ();
    set_status (_("Ok"));
    gtk_main ();
    gdk_threads_leave ();
}
